/**************************************************************************

    Written by Imran Haque (ihaque@cs.stanford.edu) based on GPL3-licensed
    code (C) 2010 by Eli Lilly and Company

    Copyright (C) 2011 Stanford University


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************/

#include "fingerprint.h"

#ifdef _WIN32
#include <windows.h>
static inline double getTimeMilliseconds(void) {
    return GetTickCount();
}
#define BENCH_OK
#elif defined(__APPLE__) || defined(__linux) || defined(UNIX)
#include <sys/time.h>
static inline double getTimeMilliseconds(void) {
    struct timeval tv;
    gettimeofday(&tv,NULL);
    return tv.tv_sec*1000.0 + tv.tv_usec/1000.0;
}
#define BENCH_OK
#else
#warning Could not detect operating system timing function; benchmarking will be unavailable.
#endif

static int choose_next_centre ( int & icentre,int * selected,int size)
{
  for(int i=0;i<size;++i)
  {
    if(! selected[i])
    {
      icentre = i;
      return 1;
    }
  }
  return 0;
}

// Selects two potential centers for speculative leader algorithm with D=2
static int two_potl_centers(int* centers,int* selected,int size) {
    int center_count = 0;
    for (int i = 0; i < size && center_count < 2; i++) {
        if (!selected[i]) {
            centers[center_count] = i;
            center_count++;
        }
    }
    return center_count;
}

static void usage(int rc)
{
  cerr << "Performs leader clustering on a set of fingerprints\n";
  cerr << "Usage <options> <input_file>\n";
  cerr << "options:\n";
  cerr << " -h             display this message\n";
  cerr << " -b             run Tanimoto distance benchmark\n";
  cerr << " -t <dis>       specify distance threshold (default 0.2)\n";
  cerr << " -A <file>      specify previously selected fingerprints\n";

  exit (rc);
}

void run_bench(const FingerprintPool& pool)
{
    #ifdef BENCH_OK
    const int tanimotos = 1 << 26;
    for (int n_fps = 1; n_fps < pool.num_fps(); n_fps <<= 1) {
        double start = getTimeMilliseconds();
        pool.benchmark(tanimotos,n_fps);
        double end = getTimeMilliseconds();
        cout << "Computing "<<tanimotos<<" Tanimotos on "<<n_fps<<" different "<<pool.fpsize()<<"b FPs took "<<(end-start)<<" ms";
        cout << " = "<<(tanimotos/(end-start))<<" kTan/sec"<<endl;
    }
    #else
    cerr << "Benchmarking unavailable in this build"<<endl;
    #endif
    return;

}

int main (int argc, char ** argv)
{

    float distance_threshold = 0.2F;
    char previously_selected_file[200];
    int previous = 0;
    int i;
    bool benchmark = false;
    for( i=1;(i<argc) && ((argv[i])[0] == '-');i++)
    {
        switch ((argv[i])[1])
        {
            case 'h': usage(0); break;
            case 'b': benchmark = true; break;
            case 't': i++; distance_threshold = atof(argv[i]); break;
            case 'A': i++; previous=1; strcpy(previously_selected_file,argv[i]); break;
            default: 
                      cerr<<"\nUnrecognized option "<< argv[i] <<endl<<endl;
                      usage(0);
        }
    }
    if(i != (argc-1) ) 
    {
        cerr << "No input fp file !"<<endl;
        usage(0);
    }

    FingerprintPool pool(argv[i]);
    int size = pool.num_fps();

    cerr <<"Pool has "<< size << " fingerprints." <<endl;
    if (benchmark) {
        run_bench(pool);
        return 0;
    }
    int *selected=new int[size];
    float* distances = new float[size];
    for( i=0;i<size;i++)
        selected[i] = 0;

    #ifdef SPECULATIVE
    bool speculative = true;
    #elif defined(NOSPECULATIVE)
    bool speculative = false;
    #else
    bool speculative = true;
    #endif

    int cluster_id = 0;
    if(previous)
    {
        FingerprintPool pool2(previously_selected_file);
        int size2 = pool2.num_fps();
        cerr <<"Previously selected Pool has "<< size2 << " fingerprints." <<endl;
        if (!speculative) {
            for(int j =0; j<size2;j++) {
                #pragma omp parallel for schedule(dynamic,1024)
                for( i=0;i<size;i++) {
                    if(selected[i]) continue;
                    float dis = pool2.tanimoto_distance(j,pool,i);
                    if(dis <= distance_threshold) selected[i] = -1;
                }
            }
        } else {
            float cand_distances[2];
            int j;
            for(j =0; j<(size2-1);j++) {
                #pragma omp parallel for schedule(dynamic,1024) private(cand_distances)
                for( i=0;i<size;i++) {
                    if(selected[i]) continue;
                    pool2.tanimoto_distance_2(j,j+1,pool,i,cand_distances);
                    if(cand_distances[0] <= distance_threshold || cand_distances[1] <= distance_threshold)
                        selected[i] = -1;
                }
            }
            if (j < size2) {
                // Clean up possible last element in pool2
                #pragma omp parallel for schedule(dynamic,1024)
                for( i=0;i<size;i++) {
                    if(selected[i]) continue;
                    float dis= pool2.tanimoto_distance(j,pool,i);
                    if(dis <= distance_threshold) selected[i] = -1;
                }
            }
        }
    }
    int icentre=0;
    int start =0 ;
    // Standard leader algorithm
    if (!speculative) {
        while (choose_next_centre (icentre,&selected[start],size - start))
        {
            cluster_id ++;
            icentre += start;
            //cout<<cluster_id <<" "<<pool.id(icentre)<<" "<<0.0F<<endl;
            start = icentre + 1;
            selected[icentre] = cluster_id;
            distances[icentre] = 0.0f;
            #pragma omp parallel for schedule(dynamic,256)
            for( i= start;i<size;++i)
            {
                if (selected[i]) continue;
                distances[i] = pool.tanimoto_distance(icentre,i);
                if (distances[i] <= distance_threshold) selected[i] = cluster_id;
            }
        }
    } 
    // Speculative leader algorithm
    else {
        // Speculative projection onto two centers to reuse data
        int potl_centers[2];
        int n_options;
        float cand_distances[2];
        while (n_options = two_potl_centers(potl_centers,selected+start,size-start)) {
            potl_centers[0] += start; potl_centers[1] += start; // Fix up offset
            if (n_options == 1 || (pool.tanimoto_distance(potl_centers[0],potl_centers[1]) <= distance_threshold)) {
                // We only have one option - either this is the last one, or the second "option" will
                // actually get assigned to the first center.
                cluster_id++;
                icentre = potl_centers[0];
                start = icentre + 1;
                //cout << "N"<<cluster_id << " " << pool.id(icentre) << " " << 0.0f<<endl;
                selected[icentre] = cluster_id;
                distances[icentre] = 0.0f;
                #pragma omp parallel for schedule(dynamic,256)
                for(int i= start;i<size;++i)
                {
                    if (!selected[i]) {
                        distances[i] = pool.tanimoto_distance(icentre,i);
                        if (distances[i] <= distance_threshold) selected[i] = cluster_id;
                    }
                }
            } else {
                // Both of these will end up being actual centers so amortize FPDB load costs by comparing
                // everything else to both
                const int c0id = cluster_id+1;
                const int c1id = cluster_id+2;
                cluster_id += 2;
                start = potl_centers[1] + 1;
                //cout << "S"<<c0id << " " << pool.id(potl_centers[0]) << " " << 0.0f<<endl;
                //cout << "S"<<c1id << " " << pool.id(potl_centers[1]) << " " << 0.0f<<endl;
                selected[potl_centers[0]] = c0id;
                selected[potl_centers[1]] = c1id;
                distances[potl_centers[0]] = 0.0f;
                distances[potl_centers[1]] = 0.0f;
                #pragma omp parallel for schedule(dynamic,256) private(cand_distances)
                for (int i = start; i < size; i++) {
                    if (!selected[i]) {
                        pool.tanimoto_distance_2(potl_centers[0],potl_centers[1],i,cand_distances);
                        if (cand_distances[0] <= distance_threshold) {
                            selected[i]  = c0id;
                            distances[i] = cand_distances[0];
                        } else if (cand_distances[1] <= distance_threshold) {
                            selected[i]  = c1id;
                            distances[i] = cand_distances[1];
                        }
                    }
                }
            }
        }
    }

    // Clustering output
    for (int i = 0; i < size; i++) {
        if (selected[i] < 0) {
            // Assigned to a previous FP pool, which are all cluster 0
            cout<<0 <<" "<<pool.id(i)<<" "<<distances[i]<<endl;
        } else {
            cout<<selected[i] <<" "<<pool.id(i)<<" "<<distances[i]<<endl;
        }
    }
    cerr << "Found " << cluster_id << " clusters"<<endl;
    delete [] selected;

    return 0;
}
