/**************************************************************************

    Written by Imran Haque (ihaque@cs.stanford.edu) based on GPL3-licensed
    code (C) 2010 by Eli Lilly and Company

    Copyright (C) 2011 Stanford University


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************/

#include "fingerprint.h"
#include <fstream>

#ifdef _DEBUG
#include <assert.h>
#else
#define assert(x)
#endif


#ifdef POPCNT_SSE4
#include "popcount_sse4.h"
#elif defined(POPCNT_SSE2_32)
#include "popcount_sse2_32.h"
#elif defined(POPCNT_SSE2_16)
#include "popcount_sse2_16.h"
#elif defined(POPCNT_SSE2_8)
#include "popcount_sse2_8.h"
#elif defined(POPCNT_SSSE3)
#include "popcount_ssse3.h"
#else
#include "popcount_lut8.h"
#endif

inline unsigned HTOI(unsigned char a) {return ((a >= 'A' )?(a - 'A' + 10 ):(a - '0'));}

void Fingerprint::clear()
{
  if (_fps != NULL) {
        if (_alignment) {
          _fps = (unsigned int*)(((intptr_t)_fps)-_alignment);
        }
        delete[] _fps;
        delete[] _npos;
        _fps = NULL;
        _npos= NULL;
        _nfp = _wordsperfp = _totalfpwords = 0;
  }
}

Fingerprint::Fingerprint() : _fps(NULL),_npos(NULL),_wordsperfp(0),_totalfpwords(0),_nfp(0),_alignment(0)
{
}

Fingerprint::~Fingerprint()
{
  clear();
}

int Fingerprint::init(string id, string hex)
{
  clear();
  _id = id;
  _nfp = count(hex.begin(),hex.end(),',') + 1;

  _wordsperfp = (hex.size()-(_nfp-1))/8;
  assert(_wordsperfp % 4 == 0); // For vectorization
  //printf("Words/FP: %d\n",_wordsperfp);
  _totalfpwords = _wordsperfp*_nfp;

  _fps  = new unsigned int[_totalfpwords+3]; // May need up to 3 words to align start
  _npos = new unsigned int[_nfp];

  int misaligned = ((intptr_t)_fps) & 0xF;
  if (misaligned) {
    _alignment = 16 - misaligned;
    _fps = (unsigned int*)(((intptr_t)_fps)+_alignment);
  } else {
        _alignment = 0;
  }

  int idx = 0;
  istringstream iss(hex);
  string pch;
  while(getline(iss,pch,','))
  {
    assert(idx < _nfp);
    assert(pch.size() == _wordsperfp*8); // hex = 4bits * 8 = 32b word
    unsigned* fp = _fps + idx*_wordsperfp;
    for (int i = 0; i < _wordsperfp; i++) {
        fp[i] = (HTOI(pch[8*i+0]) << 28) | (HTOI(pch[8*i+1]) << 24) | 
                (HTOI(pch[8*i+2]) << 20) | (HTOI(pch[8*i+3]) << 16) | 
                (HTOI(pch[8*i+4]) << 12) | (HTOI(pch[8*i+5]) << 8)  | 
                (HTOI(pch[8*i+6]) << 4)  | (HTOI(pch[8*i+7]));
    }
    _npos[idx] = popcount_1fp(fp,_wordsperfp);

  }
  return _nfp;
}

FingerprintPool::FingerprintPool(const char* fname) 
{
  ifstream fin(fname);
  int lines = count(istreambuf_iterator<char>(fin),istreambuf_iterator<char>(),'\n');
  if(lines == 0)
  {
    cerr << "There are no fingerprints in '" <<fname <<"'"<<endl;
    exit(1);
  }
  cerr << fname <<" has " << lines <<" lines." <<endl;
  _nfp = lines;
  _ids = new string[_nfp];
  _wordsperpart = _partsperfp = _alignment = 0;
  _fps = _npos = NULL;
  
  Fingerprint fpobj;
  fin.seekg(0);
  int i = 0;
  string id;
  string hex;
  while(getline(fin,id,' ') && getline(fin,hex))
  { 
    if(!fpobj.init(id,hex))
    {
      cerr<< "Failed to initialize " << id << " with fingerprint " << hex <<endl;
      exit(1);
    }
    if (_fps == NULL) {
      // Allocate arrays now
      assert(fpobj.words() % 4 == 0); // Alignment restriction
      _fps  = new unsigned[_nfp * fpobj.words() + 3]; // 3 extra to guarantee alignability
      _npos = new unsigned[_nfp * fpobj.num_fps()];
      if (!_fps || !_npos) {
        cerr<< "Failed to allocate memory for fingerprint pool" << hex <<endl;
        exit(2);
      }
      _wordsperpart = fpobj.words_per_fp();
      _partsperfp = fpobj.num_fps();
      int misaligned = ((intptr_t)_fps) & 0xF;
      if (misaligned) {
         _alignment = 16 - misaligned;
         _fps = (unsigned int*)(((intptr_t)_fps)+_alignment);
       } else {
         _alignment = 0;
       }
    }
    // Make sure this fingerprint has the right size
    if (fpobj.words_per_fp() != _wordsperpart || fpobj.num_fps() != _partsperfp) {
      cerr << "Fingerprint " <<id<<" has different length than the first one" <<endl;
      exit(3);
    }
    // Copy this fingerprint into the pool
    memcpy(_fps+i*_wordsperpart*_partsperfp,fpobj.get_fps(),fpobj.words()*sizeof(unsigned));
    memcpy(_npos+i*_partsperfp,fpobj.get_pos(),_partsperfp*sizeof(unsigned));
    _ids[i] = id;
    i++;
  }
  fin.close();
}

FingerprintPool::~FingerprintPool()
{
    if (_fps) {
        if (_alignment) {
          _fps = (unsigned int*)(((intptr_t)_fps)-_alignment);
        }
        delete[] _ids;
        delete[] _fps;
        delete[] _npos;
    }
}

float FingerprintPool::tanimoto_distance(const int idxA,const int idxB) const
{
    return tanimoto_distance(idxA,*this,idxB);
}

float FingerprintPool::tanimoto_distance(const int idxA,const FingerprintPool& B,const int idxB) const
{
    unsigned int* pa = _fps + idxA*_wordsperpart*_partsperfp;
    unsigned int* pb = B._fps + idxB*B._wordsperpart*B._partsperfp;
    unsigned int* pa_pos = _npos + idxA*_partsperfp;
    unsigned int* pb_pos = B._npos + idxB*B._partsperfp;
    assert(_partsperfp == B._partsperfp);
    assert(_wordsperpart == B._wordsperpart);

    float rc = 0.0F;
    int N = _partsperfp;
    int fplen = _wordsperpart;
    for(int i=0;i<N;++i)
    {
       int common = popcount_2fp(pa+i*fplen,pb+i*fplen,fplen);
       int sum = pa_pos[i] + pb_pos[i] - common;
       if (sum != 0) 
            rc += (float)common/((float)sum);
       else
            rc += 1.0f;
    }
    return 1.0F - rc/((float)N);
}

void FingerprintPool::tanimoto_distance_2(const int idxR0,const int idxR1,const int idxB,float* distances) const
{
    tanimoto_distance_2(idxR0,idxR1,*this,idxB,distances);
    return;
}
void FingerprintPool::tanimoto_distance_2(const int idxR0,const int idxR1,const FingerprintPool& B,const int idxB,float* distances) const
{
    unsigned int* pR0 = _fps + idxR0*_wordsperpart*_partsperfp;
    unsigned int* pR1 = _fps + idxR1*_wordsperpart*_partsperfp;
    unsigned int* pb = B._fps + idxB*B._wordsperpart*B._partsperfp;
    unsigned int* pR0_pos = _npos + idxR0*_partsperfp;
    unsigned int* pR1_pos = _npos + idxR1*_partsperfp;
    unsigned int* pb_pos = B._npos + idxB*B._partsperfp;
    assert(_partsperfp == B._partsperfp);
    assert(_wordsperpart == B._wordsperpart);

    int N = _partsperfp;
    int fplen = _wordsperpart;
    int common[2];
    int sum[2];
    float rc[2] = {0,0};
    for(int i=0;i<N;++i)
    {
       popcount_2fp_2(pR0+i*fplen,pR1+i*fplen,pb+i*fplen,fplen,common);
       sum[0] = pR0_pos[i] + pb_pos[i] - common[0];
       sum[1] = pR1_pos[i] + pb_pos[i] - common[1];
       // Add checks here to guard against 0/0 nans
       if (sum[0] != 0) rc[0] += (float)common[0]/((float)sum[0]);
       else             rc[0] += 1.0f;

       if (sum[1] != 0) rc[1] += (float)common[1]/((float)sum[1]);
       else             rc[1] += 1.0f;
    }
    distances[0] = 1.0f - rc[0]/((float)N);
    distances[1] = 1.0f - rc[1]/((float)N);
    return;
}

float FingerprintPool::benchmark(const int tanimotos_to_compute,const int fps_to_use) const 
{
    float distance = 0;
    int nfp = (fps_to_use > _nfp) ? _nfp : fps_to_use;
    // Add reduction clause to prevent this from being optimized away
    #pragma omp parallel for reduction(+:distance)
    for (int i = 0; i < tanimotos_to_compute; i++) {
        int fpidx = i % nfp;
        distance += tanimoto_distance(0,fpidx);
    }
    return distance;
}

