/**************************************************************************

    Written by Imran Haque (ihaque@cs.stanford.edu) based on GPL3-licensed
    code (C) 2010 by Eli Lilly and Company

    Copyright (C) 2011 Stanford University


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************/

#ifndef _FINGERPRINT_H
#define _FINGERPRINT_H

#include <sstream>
#include <string>
#include <iostream>
#include <cstdlib>
#include <algorithm>
#include <string.h>
using namespace std;

////////////////////////////////////////////////////
//object for single or multiple fix-length fingerprints
////////////////////////////////////////////////////////////

class Fingerprint
{
  protected:
    string _id;                     // unique identifier
    unsigned int *_fps,*_npos;     // Storage for fingerprints and positive-bit counts
    int _wordsperfp,_totalfpwords; // "Word" = 32 bits;
                                   // totalfpwords = total size, all fingerprints for this molecule
    int _nfp,_alignment;           // _nfp = # fingerprints stored here 
                                   // _alignment is internal data used for offsets to align loads

    void clear();

  private: //disable copy 
    Fingerprint(const Fingerprint &fp);
    Fingerprint& operator=(const Fingerprint &fp);

  public:
    Fingerprint();
    ~Fingerprint();

    int init(string id, string hex);

    int valid(){return _fps != NULL;}

    const char * id() {return _id.c_str();}
    
    int words() {return _totalfpwords;}

    unsigned int * get_fps() {return _fps;}
    unsigned int * get_pos() {return _npos;}

    int num_fps() const {return _nfp;}
    int words_per_fp() const {return _wordsperfp;}
};

class FingerprintPool {
    protected:
    string* _ids; //unique identifiers
    unsigned int *_fps,*_npos; //Storage for fingerprints and positive-bit counts
    int _wordsperpart,_partsperfp,_nfp,_alignment;

    public:
    // Build a FingerprintPool from a file; coalesces all Fingerprint objects into one large allocation
    FingerprintPool(const char* fname);
    ~FingerprintPool();
    
    // Compute the Tanimoto distance between two fingerprints in this pool
    float tanimoto_distance(const int idxA,const int idxB) const;

    // Compute Tanimoto distance between fingerprint idxA in this pool and idxB in a different Pool
    float tanimoto_distance(const int idxA,const FingerprintPool& B,const int idxB) const;

    // Compute Tanimoto distances between (idxR0 and idxB) and (idxR1 and idxB) in this pool
    void tanimoto_distance_2(const int idxR0,const int idxR1,const int idxB,float* distances) const;

    // Compute Tanimoto distances between (idxR0 and idxB) and (idxR1 and idxB), where idxR0 and R1 are
    // from this pool and idxB is from another pool
    void tanimoto_distance_2(const int idxR0,const int idxR1,const FingerprintPool& B,
                             const int idxB,float* distances) const;
    // Benchmark one-vs-many Tanimoto distance calculation
    float benchmark(const int tanimotos_to_compute,const int fps_to_use) const;
    const char* id(const int idx) {return _ids[idx].c_str();}
    const int fpsize(void) const {return 32*_wordsperpart*_partsperfp;}
    int num_fps() const {return _nfp;}

};

#endif

