/**************************************************************************

 Written by Imran S. Haque (ihaque@cs.stanford.edu)

 Copyright (c) 2011 Stanford University.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.
     * Neither the name of Stanford University nor the names of its contributors
       may be used to endorse or promote products derived from this software without
       specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

**************************************************************************/

__device__ int loadTile(unsigned int *fpCache,const unsigned int* fps, const unsigned int* mags, const int num,
                        const int fpSize, const int stride, const int cacheSize,const int base)
{
    const int bound = min((base + cacheSize),num);
    const int nvalid = bound-base;
    for (int i = 0; i < nvalid; i++) {
        int row = base + i;
        for (int j = 0; j < fpSize; j += blockDim.x) {
            fpCache[i*(fpSize+1) + j + threadIdx.x] = (j+threadIdx.x < fpSize) ? fps[row*stride + j + threadIdx.x] : 0;
        }
        if (threadIdx.x == 0) fpCache[i*(fpSize+1)+fpSize] = mags[row];
    }
    __syncthreads();
    return nvalid;
}

__global__ void countNeighbors1k(const unsigned int* qfps, const unsigned int* dbfps,
                                 const unsigned int* magQs, const unsigned int* magDs,
                                 const int numQ, const int numDB, const float cutoff,
                                 const int strideQ, const int strideDB, unsigned int* counts)
{
    const int qfpCacheSize = 16;
    const int l2qfpCacheSize = 4; // log2
    const int dfpCacheSize = 16;
    const int l2dfpCacheSize = 4; // log2
    const int fpSize = 32;
    const int tileSize = qfpCacheSize * dfpCacheSize;
    __shared__ unsigned int qfpCache[qfpCacheSize * (fpSize+1)]; // +1 to avoid bank conflicts and to store magnitudes
    __shared__ unsigned int dfpCache[dfpCacheSize * (fpSize+1)];
    __shared__ unsigned int tileCounts[qfpCacheSize];
    for (int qfpBase = blockIdx.x*qfpCacheSize; qfpBase < numQ; qfpBase += qfpCacheSize * gridDim.x) {
        // Load a tile of query fingerprints
        int nvalidQ = loadTile(qfpCache,qfps,magQs,numQ,fpSize,strideQ,qfpCacheSize,qfpBase);
        if (threadIdx.x < qfpCacheSize) tileCounts[threadIdx.x] = 0;
        __syncthreads();
        // Iterate over all database fingerprints
        for (int dfpBase = 0; dfpBase < numDB; dfpBase+=dfpCacheSize)
        {
            // Load database FP into smem
            int nvalidD = loadTile(dfpCache,dbfps,magDs,numDB,fpSize,strideDB,dfpCacheSize,dfpBase);
            __syncthreads();
            // Compute all Tanimotos in this tile
            for (int tanIdxB = 0; tanIdxB < tileSize; tanIdxB += blockDim.x) {
                int tanIdx = tanIdxB + threadIdx.x;
                if (tanIdx < tileSize) {
                    int qcIdx = tanIdx >> l2dfpCacheSize;
                    int dcIdx = tanIdx & ((1 << l2dfpCacheSize)-1);
                    int isct = 0;
                    const unsigned int *qfp = qfpCache + qcIdx*(fpSize+1);
                    const unsigned int *dfp = dfpCache + dcIdx*(fpSize+1);
                    const unsigned int magQ = qfp[fpSize];
                    const unsigned int magD = dfp[fpSize];
                    #pragma unroll 4
                    for (int i = 0; i < fpSize; i++) {
                        isct += __popc(qfp[i] & dfp[i]);
                    }
                    float tanimoto = __fdiv_rn((float)isct,magQ+magD-isct);
                    // To avoid warp divergence, all threads will compute Tanimotos in the
                    // tile, but only ones operating on valid DB FPs will write back results
                    //if (tanimoto > cutoff && dcIdx < nvalidD) atomicAdd(tileCounts+qcIdx,1);
                    if (tanimoto > cutoff && dcIdx < nvalidD && qcIdx < nvalidQ) atomicAdd(counts+qcIdx+qfpBase,1);
                }
            }
            __syncthreads();
        }
        __syncthreads();
        // Write out count of neighbors to gmem
        //if (threadIdx.x < nvalidQ) counts[qfpBase + threadIdx.x] = tileCounts[threadIdx.x];
        __syncthreads();
    }
}
