/**************************************************************************

 Written by Imran S. Haque (ihaque@cs.stanford.edu) and Trevor Kramer

 Copyright (c) 2011 Stanford University.
 All rights reserved.

 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.
     * Neither the name of Stanford University nor the names of its contributors
       may be used to endorse or promote products derived from this software without
       specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

**************************************************************************/

#include <iostream>
#include <fstream>
#include <string>
#include <bitset>
#include <sstream>
#include <vector>
#include <cstdlib>
#include <cassert>
#include <algorithm>
#include <stdint.h>

#include <string.h>
#include <math.h>
#include <omp.h>

#include <nmmintrin.h>
#include "nbrCounts.h"

using namespace std;

void compute_magnitudes(const unsigned *qfps,unsigned* magQs,const int numqfps,const int fpT_stride_ints)
{
    #pragma omp parallel for
    for (int i = 0; i < numqfps; i++) {
        uint64_t* qfp = (uint64_t*)(qfps + i*fpT_stride_ints);
        for (int off = 0; off < fpT_stride_ints/2; off++) {
            magQs[i] += _mm_popcnt_u64(qfp[off]);
        }
    }
    return;
}

int bitcount_32bit(unsigned int n){
    int ret=0;
    while (n!=0) {
        n&=(n-1);
        ret++;
    }
    return ret;
}
inline int bitcount_32bit_sse4(unsigned int n){
    return _mm_popcnt_u32(n);
}
inline void demorton(uint32_t *x, uint32_t *y, uint32_t m) {
    uint64_t mp = ((((uint64_t)m)&0xAAAAAAAA) << 31) | (m&0x55555555);
    mp = (mp | (mp >> 1)) & 0x3333333333333333ULL; 
    mp = (mp | (mp >> 2)) & 0x0F0F0F0F0F0F0F0FULL; 
    mp = (mp | (mp >> 4)) & 0x00FF00FF00FF00FFULL; 
    mp = (mp | (mp >> 8)) & 0x0000FFFF0000FFFFULL; 
    *x = (mp >> 32)&0xFFFFFFFF;
    *y = mp & 0xFFFFFFFF;
    return;
}
inline uint64_t extract_even_64(uint64_t mp) {
    mp &= 0x5555555555555555ULL;
    mp = (mp | (mp >> 1))  & 0x3333333333333333ULL; 
    mp = (mp | (mp >> 2))  & 0x0F0F0F0F0F0F0F0FULL; 
    mp = (mp | (mp >> 4))  & 0x00FF00FF00FF00FFULL; 
    mp = (mp | (mp >> 8))  & 0x0000FFFF0000FFFFULL; 
    mp = (mp | (mp >> 16)) & 0x00000000FFFFFFFFULL; 
    return mp;
}
inline void demorton64(uint32_t *x, uint32_t *y, uint64_t m) {
    *y = (uint32_t)(extract_even_64(m) & 0xFFFFFFFF);   
    *x = (uint32_t)(extract_even_64(m >> 1) & 0xFFFFFFFF);
}

string int32_toString(unsigned int i){
    string result;
    int maxLoops = 34;
    int loop = 0;
    while ( i > 0 ) {
        /* bitwise AND operation with the first bit */
        (i & 0x80000000) ? (result += "1") : (result += "0");
        i <<= 1;
        loop += 1;
        if (loop >= maxLoops) {
            cerr << "SOMETHING FISHY in int32_toString!" <<endl;
            exit(-1);
        }
    }
    while (result.size() < 32) {
        result += "0";
    }
    return result;
}

void printBitMatrix_padded(unsigned int* bitMatrix, int numFpsPadded){
    bool doit = true;
    if (numFpsPadded > 200) {
        doit = false;
        cerr << "WARNING: in printBitMatrix_padded(), matrix > 200x200" << endl;
        cerr << "you probably dont really want to view this - so skipping..." << endl;
    }
    unsigned int ct = 0;
    int intIdx;
    for (int colIdx=0;colIdx<numFpsPadded;colIdx++) {
        if (doit) cerr << colIdx << " ";
        for (int rowIdx=0;rowIdx<numFpsPadded/32;rowIdx++) {
            intIdx = colIdx+rowIdx*numFpsPadded; 
            if (doit) cerr << int32_toString(bitMatrix[intIdx]);
            ct += bitcount_32bit(bitMatrix[intIdx]);
        }
        if (doit) cerr << endl;
    }
    cerr << "total onBits: " << ct << endl;
}

unsigned int char32BitString_to_uint32(const char *char32){
    unsigned int num=0;
    /*if (strlen(char32) != 32) {
        fprintf(stderr, "ERROR: in char32BitString_to_uint32(), input char* not of length 32\n");
        fprintf(stderr, " %d\n",(int)strlen(char32));
        exit(-1);
    }*/   
    for (int i=31; i >= 0; i--) {
        num <<= 1;
        num |= ((*char32) == '1');
        char32++;
    }
    return num;
}


NbrCounts::NbrCounts(string databasefile, string vendorfile, float cutoff, string resultsfile ){
    int numThreads = 128;
    if (numThreads  != 32*(numThreads/32)) {
        cerr << "numThreads needs to be a multiple of 32" << endl;
        cerr << numThreads << endl;
        exit(-1);
    }
    this->databasefile = databasefile;
    this->vendorfile  = vendorfile;
    this->resultsfile = resultsfile;
    this->cutoff = cutoff; 
    this->numThreads=numThreads;

    this->verbose_gpu_matrix  = true;
    this->verbose_gpu_looping  = false;

    this->numFpsTrue  = 0; //the true number of fps
    this->numFpsTrue_vendor  = 0; //the true number of vendor fps
    this->numFpsPadded= 0; //add extra dummy fps so that numFps%32 == 0
    this->fpSizeInts = 0;

    ifstream databaseFile(this->databasefile.c_str());
    if (!databaseFile) {
        cerr << "ERROR: could not read file " << this->databasefile << endl;
        exit(-1);
    }
    ifstream vendorFile(this->vendorfile.c_str());
    if (!vendorFile) {
        cerr << "ERROR: could not read file " << this->vendorfile << endl;
        exit(-1);
    }

    ofstream outfile(this->resultsfile.c_str());
    if (!outfile) {
        cerr << "ERROR: could not write file " << this->resultsfile << endl;
        exit(-1);
    }
    outfile.close();

    // setting up so we can handle arbitrarily sized bitstrings
    this->fpBitPadLen = 0; 
    int fpBitLenRaw = 0;
    int fpBitLen = 0; // == fpBitLenRaw + fpBitPadLen  (typically 1024)
    if (sizeof(unsigned int) != 4) {
        cerr << "ERROR: only setup to handle sizeof(unsigned int) == 4 bytes" << endl;
        cerr << sizeof(unsigned int) << endl;
        exit(-1);
    }

    string line;
    while (!databaseFile.eof()) {
        getline(databaseFile, line);
        if (!databaseFile.eof()) {
            if (this->numFpsTrue == 0) {
                stringstream ss(line);
                string id;
                ss >> id;
                string fp;
                ss >> fp;
                fpBitLenRaw = fp.size();
                fpBitLen = fpBitLenRaw;
                this->fpSizeInts = fpBitLenRaw/32;
                if (32*this->fpSizeInts < fpBitLenRaw) { //non-standard fp length (like ap or tt) - so we must pad
                    this->fpSizeInts += 1;
                    fpBitLen = 32*this->fpSizeInts;
                } else if (32*this->fpSizeInts > fpBitLenRaw) { //non-standard fp length (like ap or tt) - so we must pad
                    cerr << "should not be getting here" << endl;
                    cerr << fpBitLenRaw << endl;
                    exit(-1);
                }
                this->fpBitPadLen = fpBitLen-fpBitLenRaw; //need to pad the last fp string chunk with this many zeros for 32 bit alignment
            }
            this->numFpsTrue++;
        }
    }
    databaseFile.close();

    //////// count vendor fps
    while (!vendorFile.eof()) {
        getline(vendorFile, line);
        if (!vendorFile.eof()) {
            this->numFpsTrue_vendor++;
        }
    }
    vendorFile.close();

    // handle padding database
    this->numFpsPadded = numThreads*(this->numFpsTrue/numThreads);
    if (this->numFpsPadded != this->numFpsTrue) {
        this->numFpsPadded += numThreads;
    }

    cerr << "numFpsTrue  : " << this->numFpsTrue   << endl;
    cerr << "numFpsPadded: " << this->numFpsPadded << endl;

    //////////////////// read in database fps
    databaseFile.open(this->databasefile.c_str());
    int nbytes_fp = this->fpSizeInts * 4;
    this->fingerprints = (unsigned int*) calloc(this->numFpsPadded, nbytes_fp); //will have a stride of 0      (gpu fast)
    string padding (this->fpBitPadLen,'0');
    for (int i = 0; !databaseFile.eof(); i++) {
        getline(databaseFile, line);
        if (!databaseFile.eof()) {
            stringstream ss(line);
            string id;
            ss >> id;
            this->ids.push_back (id);
            string fp;
            ss >> fp;
            fp += padding;
            //cerr << fp << endl;
            for (int j = 0; j < this->fpSizeInts; j++) {
                this->fingerprints[j * this->numFpsPadded + i] = char32BitString_to_uint32(fp.substr(32 * j, 32).c_str ());
            }
        }
    }
    databaseFile.close();

    // Set up transposed aligned fingerprints matrix - fast for CPU since there's no vectorized popcnt instruction
    this->fpT_stride_ints = ((nbytes_fp+15)/16)*4;
    // Add 16 to ensure we can align it
    this->fingerprintsT_alloc = (unsigned int*) malloc(this->numFpsTrue * fpT_stride_ints * 4 + 16);
    assert( (((uintptr_t)(this->fingerprintsT_alloc))&(0xF)) == 0);
    this->fingerprintsT = this->fingerprintsT_alloc;
    this->magD = (unsigned int*)malloc(this->numFpsTrue * 4);
    // Transpose the GPU-optimal fingerprint matrix into the CPU-optimal one
    for (int fpidx = 0; fpidx < this->numFpsTrue; fpidx++) {
        unsigned int* fp = this->fingerprintsT + fpidx * this->fpT_stride_ints;
        unsigned int mag = 0;
        for (int off = 0; off < this->fpSizeInts; off++) {
            fp[off] = this->fingerprints[off*this->numFpsPadded + fpidx];
            mag += bitcount_32bit_sse4(fp[off]);
        }
        this->magD[fpidx] = mag;
    }

}

unsigned int NbrCounts::myScreen(unsigned int* fp){
    bool verbose = false;
    if (verbose) {
        cerr << "vendor fp: " << endl;
        for (int intIdx =0;intIdx < this->fpSizeInts; intIdx++) {
            cerr << int32_toString(fp[intIdx]);
        }
        cerr << endl << endl;
    }

    unsigned int ct = 0;
    for (int fpIdx=0; fpIdx < this->numFpsPadded;fpIdx++) {
        //for (int fpIdx=0; fpIdx < this->numFpsTrue;fpIdx++) {
        double num = 0.0;
        double den = 0.0;
        if (verbose) {
            if (fpIdx < this->numFpsTrue) {
                cerr << "db Idx" << fpIdx << endl;
            }
        }
        for (int intIdx =0;intIdx < this->fpSizeInts; intIdx++) {
            num += bitcount_32bit(this->fingerprints[fpIdx + intIdx*this->numFpsPadded] & fp[intIdx]);
            den += bitcount_32bit(this->fingerprints[fpIdx + intIdx*this->numFpsPadded] | fp[intIdx]);
            if (verbose) {
                if (fpIdx < this->numFpsTrue) {
                    cerr << "intIdx: " << intIdx << endl;
                    cerr << "chunk db    : " << int32_toString(this->fingerprints[fpIdx + intIdx*this->numFpsPadded]) << endl;
                    cerr << "chunk vendor: " << int32_toString(fp[intIdx]) << endl;
                    cerr << "             and: " << bitcount_32bit(this->fingerprints[fpIdx + intIdx*this->numFpsPadded] & fp[intIdx]) << endl;
                    cerr << "             or : " << bitcount_32bit(this->fingerprints[fpIdx + intIdx*this->numFpsPadded] | fp[intIdx]) << endl;
                    cerr << endl;
                }
            }
        }
        double tani = 0.0;
        if (den != 0.0) {
            tani = num/den;
        }
        if (verbose) {
            if (fpIdx < this->numFpsTrue) {
                cerr << "fpIdx tani: " << fpIdx << " " << tani << endl;
            }
        }
        if (tani >= this->cutoff) {
            if (verbose) {
                cerr << "hit fpIdx: " << fpIdx << endl;
            }
            ct += 1;
        }
    }
    return ct;
}
unsigned int NbrCounts::myScreen_sse4(unsigned int* fp){
    bool verbose = false;
    if (verbose) {
        cerr << "vendor fp: " << endl;
        for (int intIdx =0;intIdx < this->fpSizeInts; intIdx++) {
            cerr << int32_toString(fp[intIdx]);
        }
        cerr << endl << endl;
    }

    unsigned int ct = 0;
    for (int fpIdx=0; fpIdx < this->numFpsPadded;fpIdx++) {
        //for (int fpIdx=0; fpIdx < this->numFpsTrue;fpIdx++) {
        double num = 0.0;
        double den = 0.0;
        if (verbose) {
            if (fpIdx < this->numFpsTrue) {
                cerr << "db Idx" << fpIdx << endl;
            }
        }
        for (int intIdx =0;intIdx < this->fpSizeInts; intIdx++) {
            num += bitcount_32bit_sse4(this->fingerprints[fpIdx + intIdx*this->numFpsPadded] & fp[intIdx]);
            den += bitcount_32bit_sse4(this->fingerprints[fpIdx + intIdx*this->numFpsPadded] | fp[intIdx]);
            if (verbose) {
                if (fpIdx < this->numFpsTrue) {
                    cerr << "intIdx: " << intIdx << endl;
                    cerr << "chunk db    : " << int32_toString(this->fingerprints[fpIdx + intIdx*this->numFpsPadded]) << endl;
                    cerr << "chunk vendor: " << int32_toString(fp[intIdx]) << endl;
                    cerr << "             and: " << bitcount_32bit_sse4(this->fingerprints[fpIdx + intIdx*this->numFpsPadded] & fp[intIdx]) << endl;
                    cerr << "             or : " << bitcount_32bit_sse4(this->fingerprints[fpIdx + intIdx*this->numFpsPadded] | fp[intIdx]) << endl;
                    cerr << endl;
                }
            }
        }
        double tani = 0.0;
        if (den != 0.0) {
            tani = num/den;
        }
        if (verbose) {
            if (fpIdx < this->numFpsTrue) {
                cerr << "fpIdx tani: " << fpIdx << " " << tani << endl;
            }
        }
        if (tani >= this->cutoff) {
            if (verbose) {
                cerr << "hit fpIdx: " << fpIdx << endl;
            }
            ct += 1;
        }
    }
    return ct;
}
unsigned int NbrCounts::myScreen_sse4_omp(unsigned int* fp){
    bool verbose = false;
    if (verbose) {
        cerr << "vendor fp: " << endl;
        for (int intIdx =0;intIdx < this->fpSizeInts; intIdx++) {
            cerr << int32_toString(fp[intIdx]);
        }
        cerr << endl << endl;
    }

    unsigned int ct = 0;
    #pragma omp parallel for reduction(+:ct)
    for (int fpIdx=0; fpIdx < this->numFpsPadded;fpIdx++) {
        //for (int fpIdx=0; fpIdx < this->numFpsTrue;fpIdx++) {
        double num = 0.0;
        double den = 0.0;
        if (verbose) {
            if (fpIdx < this->numFpsTrue) {
                cerr << "db Idx" << fpIdx << endl;
            }
        }
        for (int intIdx =0;intIdx < this->fpSizeInts; intIdx++) {
            num += bitcount_32bit_sse4(this->fingerprints[fpIdx + intIdx*this->numFpsPadded] & fp[intIdx]);
            den += bitcount_32bit_sse4(this->fingerprints[fpIdx + intIdx*this->numFpsPadded] | fp[intIdx]);
            if (verbose) {
                if (fpIdx < this->numFpsTrue) {
                    cerr << "intIdx: " << intIdx << endl;
                    cerr << "chunk db    : " << int32_toString(this->fingerprints[fpIdx + intIdx*this->numFpsPadded]) << endl;
                    cerr << "chunk vendor: " << int32_toString(fp[intIdx]) << endl;
                    cerr << "             and: " << bitcount_32bit_sse4(this->fingerprints[fpIdx + intIdx*this->numFpsPadded] & fp[intIdx]) << endl;
                    cerr << "             or : " << bitcount_32bit_sse4(this->fingerprints[fpIdx + intIdx*this->numFpsPadded] | fp[intIdx]) << endl;
                    cerr << endl;
                }
            }
        }
        double tani = 0.0;
        if (den != 0.0) {
            tani = num/den;
        }
        if (verbose) {
            if (fpIdx < this->numFpsTrue) {
                cerr << "fpIdx tani: " << fpIdx << " " << tani << endl;
            }
        }
        if (tani >= this->cutoff) {
            if (verbose) {
                cerr << "hit fpIdx: " << fpIdx << endl;
            }
            ct += 1;
        }
    }
    return ct;
}
unsigned int NbrCounts::myScreen_sse4_omp_T(unsigned int* qfp){
    unsigned int ct = 0;
    #pragma omp parallel for reduction(+:ct)
    for (int fpIdx = 0; fpIdx < this->numFpsTrue; fpIdx++) {
        const unsigned int* dbfp = this->fingerprintsT + fpIdx*this->fpT_stride_ints;
        unsigned int num = 0, den = 0;
        for (int offset = 0; offset < this->fpSizeInts; offset++) {
            unsigned int dbfp_o = dbfp[offset];
            unsigned int qfp_o  = qfp[offset];
            num += bitcount_32bit_sse4(dbfp_o & qfp_o);
            den += bitcount_32bit_sse4(dbfp_o | qfp_o);
        }
        double tani = (den == 0) ? 0.0 : ((double)num)/den;
        if (tani > this->cutoff) {
            ct += 1;
        }
    }
    return ct;
}

void NbrCounts::run_cpu(){
    //////////////////// read in vendor fps - these will NOT be end on as in the db fingerprints
    ifstream vendorFile;
    vendorFile.open(this->vendorfile.c_str());
    ofstream outfile(this->resultsfile.c_str());
    string line;
    int nbytes_fp_vendor = this->fpSizeInts * sizeof(unsigned int); //32 bit chunks = 4 ints
    string padding (this->fpBitPadLen,'0');
    int total = 0;
    for (int i = 0; !vendorFile.eof(); i++) {
        getline(vendorFile, line);
        if (!vendorFile.eof()) {
            stringstream ss(line);
            string id;
            ss >> id;
            string fp;
            ss >> fp;
            fp += padding;
            unsigned int* fp_v = (unsigned int*) calloc(1, nbytes_fp_vendor); 
            for (int j = 0; j < this->fpSizeInts; j++) {
                fp_v[j] = char32BitString_to_uint32(fp.substr(32 * j, 32).c_str ());
            }
            // outfile << i << " " << id << " " << this->myScreen(fp_v) << endl;    
            total += this->myScreen(fp_v);
        }
    }
    outfile.close();
    vendorFile.close();
}
void NbrCounts::read_fps(){
    //////////////////// read in vendor fps - these will NOT be end on as in the db fingerprints
    ifstream vendorFile;
    vendorFile.open(this->vendorfile.c_str());
    ofstream outfile(this->resultsfile.c_str());
    string line;
    int nbytes_fp_vendor = this->fpSizeInts * sizeof(unsigned int); //32 bit chunks = 4 ints
    string padding (this->fpBitPadLen,'0');
    int total = 0;
    for (int i = 0; !vendorFile.eof(); i++) {
        getline(vendorFile, line);
        if (!vendorFile.eof()) {
            stringstream ss(line);
            string id;
            ss >> id;
            string fp;
            ss >> fp;
            fp += padding;
            unsigned int* fp_v = (unsigned int*) calloc(1, nbytes_fp_vendor); 
            for (int j = 0; j < this->fpSizeInts; j++) {
                fp_v[j]=char32BitString_to_uint32(fp.substr(32 * j, 32).c_str ());
            }
        }
    }
    outfile.close();
    vendorFile.close();
}
void NbrCounts::run_cpu_sse4(){
    //////////////////// read in vendor fps - these will NOT be end on as in the db fingerprints
    ifstream vendorFile;
    vendorFile.open(this->vendorfile.c_str());
    ofstream outfile(this->resultsfile.c_str());
    string line;
    int nbytes_fp_vendor = this->fpSizeInts * sizeof(unsigned int); //32 bit chunks = 4 ints
    string padding (this->fpBitPadLen,'0');
    int total = 0;
    for (int i = 0; !vendorFile.eof(); i++) {
        getline(vendorFile, line);
        if (!vendorFile.eof()) {
            stringstream ss(line);
            string id;
            ss >> id;
            string fp;
            ss >> fp;
            fp += padding;
            unsigned int* fp_v = (unsigned int*) calloc(1, nbytes_fp_vendor); 
            for (int j = 0; j < this->fpSizeInts; j++) {
                fp_v[j] = char32BitString_to_uint32(fp.substr(32 * j, 32).c_str ());
            }
            // outfile << i << " " << id << " " << this->myScreen_sse4(fp_v) << endl;    
            total += this->myScreen_sse4(fp_v);
        }
    }
    outfile.close();
    vendorFile.close();
}
void NbrCounts::run_cpu_sse4_omp(){
    //////////////////// read in vendor fps - these will NOT be end on as in the db fingerprints
    ifstream vendorFile;
    vendorFile.open(this->vendorfile.c_str());
    ofstream outfile(this->resultsfile.c_str());
    string line;
    int nbytes_fp_vendor = this->fpSizeInts * sizeof(unsigned int); //32 bit chunks = 4 ints
    string padding (this->fpBitPadLen,'0');
    int total = 0;
    for (int i = 0; !vendorFile.eof(); i++) {
        getline(vendorFile, line);
        if (!vendorFile.eof()) {
            stringstream ss(line);
            string id;
            ss >> id;
            string fp;
            ss >> fp;
            fp += padding;
            unsigned int* fp_v = (unsigned int*) calloc(1, nbytes_fp_vendor); 
            for (int j = 0; j < this->fpSizeInts; j++) {
                fp_v[j] = char32BitString_to_uint32(fp.substr(32 * j, 32).c_str ());
            }
            // outfile << i << " " << id << " " << this->myScreen_sse4_omp(fp_v) << endl;    
            total += this->myScreen_sse4_omp(fp_v);
        }
    }
    outfile.close();
    vendorFile.close();
}
void NbrCounts::run_cpu_sse4_omp_T(){
    //////////////////// read in vendor fps - these will NOT be end on as in the db fingerprints
    ifstream vendorFile;
    vendorFile.open(this->vendorfile.c_str());
    ofstream outfile(this->resultsfile.c_str());
    string line;
    int nbytes_fp_vendor = this->fpSizeInts * sizeof(unsigned int); //32 bit chunks = 4 ints
    string padding (this->fpBitPadLen,'0');
    int total = 0;
    for (int i = 0; !vendorFile.eof(); i++) {
        getline(vendorFile, line);
        if (!vendorFile.eof()) {
            stringstream ss(line);
            string id;
            ss >> id;
            string fp;
            ss >> fp;
            fp += padding;
            unsigned int* fp_v = (unsigned int*) calloc(1, nbytes_fp_vendor); 
            for (int j = 0; j < this->fpSizeInts; j++) {
                fp_v[j] = char32BitString_to_uint32(fp.substr(32 * j, 32).c_str ());
            }
            // outfile << i << " " << id << " " << this->myScreen_sse4_omp_T(fp_v) << endl;    
            total += myScreen_sse4_omp_T(fp_v);
        }
    }
    outfile.close();
    vendorFile.close();
}
void NbrCounts::run_cuda(){
    //////////////////// read in vendor fps - these will NOT be end on as in the db fingerprints
    ifstream vendorFile;
    vendorFile.open(this->vendorfile.c_str());
    string line;
    int nbytes_fp_vendor = this->fpSizeInts * sizeof(unsigned int); //32 bit chunks = 4 ints
    string padding (this->fpBitPadLen,'0');
    //const unsigned int* dbfp4 = this->fingerprintsT + (fpIdx+3)*this->fpT_stride_ints;
    unsigned int blocksize = 128;
    unsigned int numvendorFPs = 0;
    unsigned int *vendorFPblock = (unsigned int*)malloc(blocksize*this->fpT_stride_ints*sizeof(unsigned int));
    vector<string> ids;
    for (int i = 0; !vendorFile.eof(); i++) {
        getline(vendorFile, line);
        if (!vendorFile.eof()) {
            stringstream ss(line);
            string id;
            ss >> id;
            ids.push_back(id);
            string fp;
            ss >> fp;
            fp += padding;
            // Check to see if we need to grow the buffer
            if (numvendorFPs == blocksize) {
                unsigned newblocksize = blocksize*2;
                unsigned int* temp = (unsigned int*)malloc(newblocksize*this->fpT_stride_ints*sizeof(unsigned int));
                memcpy(temp,vendorFPblock,blocksize*this->fpT_stride_ints*sizeof(unsigned int));
                free(vendorFPblock);
                vendorFPblock = temp;
                blocksize = newblocksize;
            }
            unsigned int* fp_v = vendorFPblock + numvendorFPs*this->fpT_stride_ints;
            for (int j = 0; j < this->fpSizeInts; j++) {
                fp_v[j] = char32BitString_to_uint32(fp.substr(32 * j, 32).c_str ());
            }
            numvendorFPs++;
        }
    }
    vendorFile.close();
    unsigned int* counts = (unsigned int*)malloc(numvendorFPs*sizeof(unsigned int));
    memset(counts,0,numvendorFPs*sizeof(unsigned int));
    myScreen_cuda(vendorFPblock,numvendorFPs,counts);
    ofstream outfile(this->resultsfile.c_str());
    for (int i = 0; i < numvendorFPs; i++) {
        string id = ids[i];
        outfile << i << " " << id << " " << counts[i] << endl;
    }
    outfile.close();
    free(vendorFPblock);
    free(counts);
}
void NbrCounts::run_cuda_unroll4(){
    //////////////////// read in vendor fps - these will NOT be end on as in the db fingerprints
    ifstream vendorFile;
    vendorFile.open(this->vendorfile.c_str());
    string line;
    int nbytes_fp_vendor = this->fpSizeInts * sizeof(unsigned int); //32 bit chunks = 4 ints
    string padding (this->fpBitPadLen,'0');
    //const unsigned int* dbfp4 = this->fingerprintsT + (fpIdx+3)*this->fpT_stride_ints;
    unsigned int blocksize = 128;
    unsigned int numvendorFPs = 0;
    unsigned int *vendorFPblock = (unsigned int*)malloc(blocksize*this->fpT_stride_ints*sizeof(unsigned int));
    vector<string> ids;
    for (int i = 0; !vendorFile.eof(); i++) {
        getline(vendorFile, line);
        if (!vendorFile.eof()) {
            stringstream ss(line);
            string id;
            ss >> id;
            ids.push_back(id);
            string fp;
            ss >> fp;
            fp += padding;
            // Check to see if we need to grow the buffer
            if (numvendorFPs == blocksize) {
                unsigned newblocksize = blocksize*2;
                unsigned int* temp = (unsigned int*)malloc(newblocksize*this->fpT_stride_ints*sizeof(unsigned int));
                memcpy(temp,vendorFPblock,blocksize*this->fpT_stride_ints*sizeof(unsigned int));
                free(vendorFPblock);
                vendorFPblock = temp;
                blocksize = newblocksize;
            }
            unsigned int* fp_v = vendorFPblock + numvendorFPs*this->fpT_stride_ints;
            for (int j = 0; j < this->fpSizeInts; j++) {
                fp_v[j] = char32BitString_to_uint32(fp.substr(32 * j, 32).c_str ());
            }
            numvendorFPs++;
        }
    }
    vendorFile.close();
    unsigned int* counts = (unsigned int*)malloc(numvendorFPs*sizeof(unsigned int));
    memset(counts,0,numvendorFPs*sizeof(unsigned int));
    myScreen_cuda_unroll4(vendorFPblock,numvendorFPs,counts);
    ofstream outfile(this->resultsfile.c_str());
    for (int i = 0; i < numvendorFPs; i++) {
        string id = ids[i];
        outfile << i << " " << id << " " << counts[i] << endl;
    }
    outfile.close();
    free(vendorFPblock);
    free(counts);
}

NbrCounts::~NbrCounts(){
    if (this->fingerprints) {
        free(this->fingerprints);
    }
}

#define MORTON_TILE_SIZE 64

unsigned int NbrCounts::myScreen_sse4_omp_vecT(unsigned int* qfp){
    unsigned int ct = 0;
    const int unroll_factor = 1;
    const int vector_width  = 4;
    assert(fpSizeInts % (vector_width*unroll_factor) == 0);
    const int niters = fpSizeInts/(vector_width);
    const __m128i* qfp128 = (const __m128i*) qfp;
    #pragma omp parallel for reduction(+:ct)
    for (int fpIdx = 0; fpIdx < this->numFpsTrue; fpIdx++) {
        const unsigned int* dbfp = this->fingerprintsT + fpIdx*this->fpT_stride_ints;
        unsigned int num = 0, den = 0;
        // Use __m128i to handle 4 FP ints at a time
        const __m128i* dbfp128 = (const __m128i*)dbfp;
        __m128i q, db, isct, un;
        long long isct_hi, isct_lo, un_lo, un_hi;
        for (int iter = 0; iter < niters; iter+=unroll_factor) {
            q   = _mm_load_si128( qfp128 + iter);
            db  = _mm_load_si128(dbfp128 + iter);

            isct  = _mm_and_si128(q,db);
            un    = _mm_or_si128(q,db);

            isct_lo  = _mm_extract_epi64(isct,0);
            isct_hi  = _mm_extract_epi64(isct,1);
            un_lo    = _mm_extract_epi64(un,0);
            un_hi    = _mm_extract_epi64(un,1);

            num += _mm_popcnt_u64(isct_lo)  + _mm_popcnt_u64(isct_hi);
            den += _mm_popcnt_u64(un_lo)    + _mm_popcnt_u64(un_hi);
        }

        double tani = (den == 0) ? 0.0 : ((double)num)/den;
        if (tani > this->cutoff) {
            ct += 1;
        }
    }
    return ct;
}
unsigned int NbrCounts::myScreen_sse4_omp_vecT_mag(unsigned int* qfp){
    unsigned int ct = 0;
    const int unroll_factor = 1;
    const int vector_width  = 4;
    assert(fpSizeInts % (vector_width*unroll_factor) == 0);
    const int niters = fpSizeInts/(vector_width);
    const __m128i* qfp128 = (const __m128i*) qfp;
    unsigned int magQ = 0;
    for (int off = 0; off < this->fpSizeInts; off++) {
        magQ += bitcount_32bit_sse4(qfp[off]);
    }
    #pragma omp parallel for reduction(+:ct)
    for (int fpIdx = 0; fpIdx < this->numFpsTrue; fpIdx++) {
        const unsigned int* dbfp = this->fingerprintsT + fpIdx*this->fpT_stride_ints;
        //unsigned int num = 0, den = 0;
        uint64_t num_lo = 0, num_hi = 0, den = 0;
        // Use __m128i to handle 4 FP ints at a time
        const __m128i* dbfp128 = (const __m128i*)dbfp;
        __m128i q, db, isct;
        long long isct_hi, isct_lo;
        for (int iter = 0; iter < niters; iter+=unroll_factor) {
            q   = _mm_load_si128( qfp128 + iter);
            db  = _mm_load_si128(dbfp128 + iter);

            isct  = _mm_and_si128(q,db);

            isct_lo  = _mm_extract_epi64(isct,0);
            isct_hi  = _mm_extract_epi64(isct,1);

            //num += _mm_popcnt_u64(isct_lo)  + _mm_popcnt_u64(isct_hi);
            num_lo += _mm_popcnt_u64(isct_lo);
            num_hi += _mm_popcnt_u64(isct_hi);
        }
        uint64_t num = num_lo + num_hi;
        den = magQ + magD[fpIdx] - num;

        double tani = (den == 0) ? 0.0 : ((double)num)/den;
        if (tani > this->cutoff) {
            ct += 1;
        }
    }
    return ct;
}
void NbrCounts::myScreen_sse4_omp_vecT_mag_block(unsigned int* qfps,const int numqfps,unsigned int* counts){
    const int numdbfps = this->numFpsTrue;
    unsigned int* dbfps = this->fingerprintsT;

    unsigned int* magQs = (unsigned int*)malloc(numqfps*4);
    memset(magQs,0,numqfps*4);
    #pragma omp parallel for
    for (int i = 0; i < numqfps; i++) {
        unsigned int* qfp = qfps + i*this->fpT_stride_ints;
        for (int off = 0; off < this->fpSizeInts; off++) {
            magQs[i] += bitcount_32bit_sse4(qfp[off]);
        }
    }

    unsigned maxfps = (numqfps > numdbfps) ? numqfps : numdbfps;
    const int tileSize = MORTON_TILE_SIZE;
    const int vector_width  = 4;
    assert(numqfps  % tileSize == 0);
    assert(numdbfps % tileSize == 0);
    assert(fpSizeInts % vector_width == 0);
    const int niters = fpSizeInts/vector_width;
    
    uint64_t numTiles = ((uint64_t)(maxfps/tileSize))*(maxfps/tileSize);
    unsigned int** allLocalCounts;
    const bool morton = true;
    #pragma omp parallel
    {
        #pragma omp single
        {
            allLocalCounts = new unsigned int*[omp_get_num_threads()];
        }
        int threadIdx = omp_get_thread_num();
        allLocalCounts[threadIdx] = new unsigned int[numqfps]; 
        unsigned int* localCounts = allLocalCounts[threadIdx];
        memset(localCounts,0,numqfps*4);
        #pragma omp for schedule (static,64)
        for (uint64_t tileid = 0; tileid < numTiles; tileid++) {
            // Figure out the coordinates for the tile we're operating on
            uint32_t qtile_idx, dbtile_idx;
            if (morton) {
                demorton64(&qtile_idx,&dbtile_idx,tileid);
            } else {
                qtile_idx = tileid/(numdbfps/tileSize);
                dbtile_idx = tileid%(numdbfps/tileSize);
            }
            int qfpBase  =  qtile_idx * tileSize;
            int dbfpBase = dbtile_idx * tileSize;
            if (qfpBase >= numqfps || dbfpBase >= numdbfps) {
            	//cout << "Thread "<<omp_get_thread_num() << " skipping tile "<<tileid<<": query "<<qfpBase<<":"<<(qfpBase+tileSize)<<" db "<<dbfpBase<<":"<<(dbfpBase+tileSize)<<endl;
		continue;
		}
            //cout << "Thread "<<omp_get_thread_num() << " working on tile "<<tileid<<": query "<<qfpBase<<":"<<(qfpBase+tileSize)<<" db "<<dbfpBase<<":"<<(dbfpBase+tileSize)<<endl;
            for (int qfpOff = 0; qfpOff < tileSize; qfpOff++) {
                unsigned int qfpIdx = qfpBase + qfpOff;
                unsigned int magQ = magQs[qfpIdx];
                unsigned int* qfp = qfps + (qfpIdx)*this->fpT_stride_ints;
                const __m128i* qfp128 = (const __m128i*) qfp;
                for (int dbfpOff = 0; dbfpOff < tileSize; dbfpOff++) {
                    unsigned int dbfpIdx = dbfpBase + dbfpOff;
                    const unsigned int* dbfp = dbfps + (dbfpIdx)*this->fpT_stride_ints;
                    unsigned int num = 0, den = 0;
                    // Use __m128i to handle 4 FP ints at a time
                    const __m128i* dbfp128 = (const __m128i*)dbfp;
                    __m128i q, db, isct;
                    long long isct_hi, isct_lo;
                    for (int iter = 0; iter < niters; iter++) {
                        q   = _mm_load_si128( qfp128 + iter);
                        db  = _mm_load_si128(dbfp128 + iter);
            
                        isct  = _mm_and_si128(q,db);

                        isct_lo  = _mm_extract_epi64(isct,0);
                        isct_hi  = _mm_extract_epi64(isct,1);

                        num += _mm_popcnt_u64(isct_lo)  + _mm_popcnt_u64(isct_hi);
                    }
                    den = magQ + magD[dbfpIdx] - num;

                    double tani = (den == 0) ? 0.0 : ((double)num)/den;
                    if (tani > this->cutoff) localCounts[qfpIdx] += 1;
                }
            }
        }
        #pragma omp single
        {
            // Reduce and free
            for (int t = 0; t < omp_get_num_threads(); t++) {
                unsigned int *lC = allLocalCounts[t];
                for (int i = 0; i < numqfps; i++) counts[i] += lC[i];
                delete[] lC;
            }
            delete[] allLocalCounts;
        }
    }
    free(magQs);
    return;
}
unsigned int NbrCounts::myScreen_sse4_omp_mag(unsigned int* qfp){
    unsigned int ct = 0;
#define SOM_UF 4
    const int unroll_factor = SOM_UF;
    const int vector_width  = 2;
    assert(fpSizeInts % (vector_width*unroll_factor) == 0);
    const int niters = fpSizeInts/(vector_width);
    const uint64_t* qfp64 = (const uint64_t*) qfp;
    unsigned int magQ = 0;
    for (int off = 0; off < niters; off++) {
        magQ += _mm_popcnt_u64(qfp64[off]);
    }
    #pragma omp parallel for reduction(+:ct)
    for (int fpIdx = 0; fpIdx < this->numFpsTrue; fpIdx++) {
        const unsigned int* dbfp = this->fingerprintsT + fpIdx*this->fpT_stride_ints;
        //unsigned int num = 0, den = 0;
#if SOM_UF==4
        uint64_t num = 0, num_lo = 0, num_hi = 0, den = 0;
        uint64_t num_lo2 = 0, num_hi2 = 0;
        const uint64_t* dbfp64 = (const uint64_t*)dbfp;
        uint64_t q, db, q2, db2;
        uint64_t q3, db3, q4, db4;
        uint64_t isct_hi, isct_lo;
        uint64_t isct_hi2, isct_lo2;
        for (int iter = 0; iter < niters; iter+=unroll_factor) {
            q   =  qfp64[iter];
            db  = dbfp64[iter];
            q2  =  qfp64[iter+1];
            db2 = dbfp64[iter+1];
            q3  =  qfp64[iter+2];
            db3 = dbfp64[iter+2];
            q4  =  qfp64[iter+3];
            db4 = dbfp64[iter+3];
            isct_lo = q & db;
            isct_hi = q2 & db2;
            isct_lo2 = q3 & db3;
            isct_hi2 = q4 & db4;
            num_lo += _mm_popcnt_u64(isct_lo);
            num_hi += _mm_popcnt_u64(isct_hi);
            num_lo2 += _mm_popcnt_u64(isct_lo2);
            num_hi2 += _mm_popcnt_u64(isct_hi2);
        }
        num = num_lo + num_hi;
        num += num_lo2 + num_hi2;
        #elif SOM_UF == 2
        uint64_t num = 0, num_lo = 0, num_hi = 0, den = 0;
        const uint64_t* dbfp64 = (const uint64_t*)dbfp;
        uint64_t q, db, q2, db2;
        uint64_t isct_hi, isct_lo;
        for (int iter = 0; iter < niters; iter+=unroll_factor) {
            q   =  qfp64[iter];
            db  = dbfp64[iter];
            q2  =  qfp64[iter+1];
            db2 = dbfp64[iter+1];
            isct_lo = q & db;
            isct_hi = q2 & db2;
            num_lo += _mm_popcnt_u64(isct_lo);
            num_hi += _mm_popcnt_u64(isct_hi);
        }
        num = num_lo + num_hi;
#endif
        den = magQ + magD[fpIdx] - num;

        double tani = (den == 0) ? 0.0 : ((double)num)/den;
        if (tani > this->cutoff) {
            ct += 1;
        }
    }
    return ct;
}
void NbrCounts::myScreen_sse4_omp_mag_block(unsigned int* qfps,const int numqfps,unsigned int* counts, unsigned int* magQs){
    #define OMB_UF 4
    const int numdbfps = this->numFpsTrue;
    uint64_t* dbfps64 = (uint64_t*)(this->fingerprintsT);
    uint64_t* qfps64 = (uint64_t*)qfps;
    const int fpT_stride_i64 = this->fpT_stride_ints/2;
    unsigned maxfps = (numqfps > numdbfps) ? numqfps : numdbfps;
    const int tileSize = MORTON_TILE_SIZE;
    const int vector_width  = 2;
    const int unroll_factor = OMB_UF;
    assert(numqfps  % tileSize == 0);
    assert(numdbfps % tileSize == 0);
    assert(fpSizeInts % (vector_width*unroll_factor) == 0);
    const int niters = fpSizeInts/(vector_width);

    bool mustAllocate = (magQs == NULL);
    if (mustAllocate) {
        magQs = (unsigned int*)malloc(numqfps*4);
        memset(magQs,0,numqfps*4);
        #pragma omp parallel for
        for (int i = 0; i < numqfps; i++) {
            uint64_t* qfp = qfps64 + i*fpT_stride_i64;
            for (int off = 0; off < niters; off++) {
                magQs[i] += _mm_popcnt_u64(qfp[off]);
            }
        }
    }

    
    uint64_t numTiles = ((uint64_t)(maxfps/tileSize))*(maxfps/tileSize);
    unsigned int** allLocalCounts;
    const bool morton = true;
    #pragma omp parallel
    {
        #pragma omp single
        {
            allLocalCounts = new unsigned int*[omp_get_num_threads()];
        }
        int threadIdx = omp_get_thread_num();
        allLocalCounts[threadIdx] = new unsigned int[numqfps]; 
        unsigned int* localCounts = allLocalCounts[threadIdx];
        memset(localCounts,0,numqfps*4);
        #pragma omp for schedule (static,64)
        for (uint64_t tileid = 0; tileid < numTiles; tileid++) {
            // Figure out the coordinates for the tile we're operating on
            uint32_t qtile_idx, dbtile_idx;
            if (morton) {
                demorton64(&qtile_idx,&dbtile_idx,tileid);
            } else {
                qtile_idx = tileid/(numdbfps/tileSize);
                dbtile_idx = tileid%(numdbfps/tileSize);
            }
            int qfpBase  =  qtile_idx * tileSize;
            int dbfpBase = dbtile_idx * tileSize;
            if (qfpBase >= numqfps || dbfpBase >= numdbfps) continue;
            for (int qfpOff = 0; qfpOff < tileSize; qfpOff++) {
                unsigned int qfpIdx = qfpBase + qfpOff;
                unsigned int magQ = magQs[qfpIdx];
                const uint64_t* qfp64 = qfps64 + (qfpIdx)*fpT_stride_i64;
                for (int dbfpOff = 0; dbfpOff < tileSize; dbfpOff++) {
                    unsigned int dbfpIdx = dbfpBase + dbfpOff;
                    const uint64_t* dbfp64 = dbfps64 + (dbfpIdx)*fpT_stride_i64;

                    #if OMB_UF==4
                    uint64_t num = 0, num_lo = 0, num_hi = 0, den = 0;
                    uint64_t num_lo2 = 0, num_hi2 = 0;
                    uint64_t q, q2, q3, q4;
                    uint64_t db, db2, db3, db4;
                    uint64_t isct_hi, isct_lo;
                    uint64_t isct_hi2, isct_lo2;
                    for (int iter = 0; iter < niters; iter+=unroll_factor) {
                        q   =  qfp64[iter];
                        db  = dbfp64[iter];
                        q2  =  qfp64[iter+1];
                        db2 = dbfp64[iter+1];
                        q3  =  qfp64[iter+2];
                        db3 = dbfp64[iter+2];
                        q4  =  qfp64[iter+3];
                        db4 = dbfp64[iter+3];
                        isct_lo = q & db;
                        isct_hi = q2 & db2;
                        isct_lo2 = q3 & db3;
                        isct_hi2 = q4 & db4;
                        num_lo += _mm_popcnt_u64(isct_lo);
                        num_hi += _mm_popcnt_u64(isct_hi);
                        num_lo2 += _mm_popcnt_u64(isct_lo2);
                        num_hi2 += _mm_popcnt_u64(isct_hi2);
                    }
                    num = (num_lo + num_hi) + (num_lo2 + num_hi2);
                    #elif OMB_UF == 2
                    uint64_t num = 0, num_lo = 0, num_hi = 0, den = 0;
                    uint64_t q, db, q2, db2;
                    uint64_t isct_hi, isct_lo;
                    for (int iter = 0; iter < niters; iter+=unroll_factor) {
                        q   =  qfp64[iter];
                        db  = dbfp64[iter];
                        q2  =  qfp64[iter+1];
                        db2 = dbfp64[iter+1];
                        isct_lo = q & db;
                        isct_hi = q2 & db2;
                        num_lo += _mm_popcnt_u64(isct_lo);
                        num_hi += _mm_popcnt_u64(isct_hi);
                    }
                    num = num_lo + num_hi;
                    #elif OMB_UF == 1
                    uint64_t num = 0, den = 0;
                    for (int iter = 0; iter < niters; iter+=unroll_factor) {
                        num += _mm_popcnt_u64( qfp64[iter] & dbfp64[iter]);
                    }
                    #endif
                    den = magQ + magD[dbfpIdx] - num;

                    double tani = (den == 0) ? 0.0 : ((double)num)/den;
                    if (tani >= this->cutoff) localCounts[qfpIdx] += 1;
                }
            }
        }
        #pragma omp single
        {
            // Reduce and free
            for (int t = 0; t < omp_get_num_threads(); t++) {
                unsigned int *lC = allLocalCounts[t];
                for (int i = 0; i < numqfps; i++) counts[i] += lC[i];
                delete[] lC;
            }
            delete[] allLocalCounts;
        }
    }
    if (mustAllocate) free(magQs);
    return;
}

void NbrCounts::run_cpu_sse4_omp_vecT(){
    //////////////////// read in vendor fps - these will NOT be end on as in the db fingerprints
    ifstream vendorFile;
    vendorFile.open(this->vendorfile.c_str());
    ofstream outfile(this->resultsfile.c_str());
    string line;
    int nbytes_fp_vendor = this->fpSizeInts * sizeof(unsigned int); //32 bit chunks = 4 ints
    string padding (this->fpBitPadLen,'0');
    int total = 0;
    for (int i = 0; !vendorFile.eof(); i++) {
        getline(vendorFile, line);
        if (!vendorFile.eof()) {
            stringstream ss(line);
            string id;
            ss >> id;
            string fp;
            ss >> fp;
            fp += padding;
            unsigned int* fp_v = (unsigned int*) calloc(1, nbytes_fp_vendor); 
            for (int j = 0; j < this->fpSizeInts; j++) {
                fp_v[j] = char32BitString_to_uint32(fp.substr(32 * j, 32).c_str ());
            }
            // outfile << i << " " << id << " " << this->myScreen_sse4_omp_vecT(fp_v) << endl;    
            total += myScreen_sse4_omp_vecT(fp_v);
        }
    }
    outfile.close();
    vendorFile.close();
}
void NbrCounts::run_cpu_sse4_omp_vecT_mag(){
    //////////////////// read in vendor fps - these will NOT be end on as in the db fingerprints
    ifstream vendorFile;
    vendorFile.open(this->vendorfile.c_str());
    ofstream outfile(this->resultsfile.c_str());
    string line;
    int nbytes_fp_vendor = this->fpSizeInts * sizeof(unsigned int); //32 bit chunks = 4 ints
    string padding (this->fpBitPadLen,'0');
    int total = 0; 
    for (int i = 0; !vendorFile.eof(); i++) {
        getline(vendorFile, line);
        if (!vendorFile.eof()) {
            stringstream ss(line);
            string id;
            ss >> id;
            string fp;
            ss >> fp;
            fp += padding;
            unsigned int* fp_v = (unsigned int*) calloc(1, nbytes_fp_vendor); 
            for (int j = 0; j < this->fpSizeInts; j++) {
                fp_v[j] = char32BitString_to_uint32(fp.substr(32 * j, 32).c_str ());
            }
            total += this->myScreen_sse4_omp_vecT_mag(fp_v);
            // outfile << i << " " << id << " " << this->myScreen_sse4_omp_vecT_mag(fp_v) << endl;    
        }
    }
    outfile.close();
    vendorFile.close();
}
void NbrCounts::run_cpu_sse4_omp_mag(){
    //////////////////// read in vendor fps - these will NOT be end on as in the db fingerprints
    ifstream vendorFile;
    vendorFile.open(this->vendorfile.c_str());
    ofstream outfile(this->resultsfile.c_str());
    string line;
    int nbytes_fp_vendor = this->fpSizeInts * sizeof(unsigned int); //32 bit chunks = 4 ints
    string padding (this->fpBitPadLen,'0');
    int total;
    for (int i = 0; !vendorFile.eof(); i++) {
        getline(vendorFile, line);
        if (!vendorFile.eof()) {
            stringstream ss(line);
            string id;
            ss >> id;
            string fp;
            ss >> fp;
            fp += padding;
            unsigned int* fp_v = (unsigned int*) calloc(1, nbytes_fp_vendor); 
            for (int j = 0; j < this->fpSizeInts; j++) {
                fp_v[j] = char32BitString_to_uint32(fp.substr(32 * j, 32).c_str ());
            }
            total += this->myScreen_sse4_omp_mag(fp_v);
            // outfile << i << " " << id << " " << this->myScreen_sse4_omp_mag(fp_v) << endl;    
        }
    }
    outfile.close();
    vendorFile.close();
}
void NbrCounts::run_cpu_sse4_omp_vecT_mag_block(){
    //////////////////// read in vendor fps - these will NOT be end on as in the db fingerprints
    ifstream vendorFile;
    vendorFile.open(this->vendorfile.c_str());
    string line;
    int nbytes_fp_vendor = this->fpSizeInts * sizeof(unsigned int); //32 bit chunks = 4 ints
    string padding (this->fpBitPadLen,'0');
    //const unsigned int* dbfp4 = this->fingerprintsT + (fpIdx+3)*this->fpT_stride_ints;
    unsigned int blocksize = 128;
    unsigned int numvendorFPs = 0;
    unsigned int *vendorFPblock = (unsigned int*)malloc(blocksize*this->fpT_stride_ints*sizeof(unsigned int));
    vector<string> ids;
    for (int i = 0; !vendorFile.eof(); i++) {
        getline(vendorFile, line);
        if (!vendorFile.eof()) {
            stringstream ss(line);
            string id;
            ss >> id;
            ids.push_back(id);
            string fp;
            ss >> fp;
            fp += padding;
            // Check to see if we need to grow the buffer
            if (numvendorFPs == blocksize) {
                unsigned newblocksize = blocksize*2;
                unsigned int* temp = (unsigned int*)malloc(newblocksize*this->fpT_stride_ints*sizeof(unsigned int));
                memcpy(temp,vendorFPblock,blocksize*this->fpT_stride_ints*sizeof(unsigned int));
                free(vendorFPblock);
                vendorFPblock = temp;
                blocksize = newblocksize;
            }
            unsigned int* fp_v = vendorFPblock + numvendorFPs*this->fpT_stride_ints;
            for (int j = 0; j < this->fpSizeInts; j++) {
                fp_v[j] = char32BitString_to_uint32(fp.substr(32 * j, 32).c_str ());
            }
            numvendorFPs++;
        }
    }
    vendorFile.close();
    unsigned int* counts = (unsigned int*)malloc(numvendorFPs*sizeof(unsigned int));
    memset(counts,0,numvendorFPs*sizeof(unsigned int));
    myScreen_sse4_omp_vecT_mag_block(vendorFPblock,numvendorFPs,counts);
    ofstream outfile(this->resultsfile.c_str());
    for (int i = 0; i < numvendorFPs; i++) {
        string id = ids[i];
        //outfile << i << " " << id << " " << counts[i] << endl;
    }
    outfile.close();
    free(vendorFPblock);
    free(counts);
}
void NbrCounts::run_cpu_sse4_omp_mag_block(){
    //////////////////// read in vendor fps - these will NOT be end on as in the db fingerprints
    ifstream vendorFile;
    vendorFile.open(this->vendorfile.c_str());
    string line;
    int nbytes_fp_vendor = this->fpSizeInts * sizeof(unsigned int); //32 bit chunks = 4 ints
    string padding (this->fpBitPadLen,'0');
    //const unsigned int* dbfp4 = this->fingerprintsT + (fpIdx+3)*this->fpT_stride_ints;
    unsigned int blocksize = 128;
    unsigned int numvendorFPs = 0;
    unsigned int *vendorFPblock = (unsigned int*)malloc(blocksize*this->fpT_stride_ints*sizeof(unsigned int));
    vector<string> ids;
    for (int i = 0; !vendorFile.eof(); i++) {
        getline(vendorFile, line);
        if (!vendorFile.eof()) {
            stringstream ss(line);
            string id;
            ss >> id;
            ids.push_back(id);
            string fp;
            ss >> fp;
            fp += padding;
            // Check to see if we need to grow the buffer
            if (numvendorFPs == blocksize) {
                unsigned newblocksize = blocksize*2;
                unsigned int* temp = (unsigned int*)malloc(newblocksize*this->fpT_stride_ints*sizeof(unsigned int));
                memcpy(temp,vendorFPblock,blocksize*this->fpT_stride_ints*sizeof(unsigned int));
                free(vendorFPblock);
                vendorFPblock = temp;
                blocksize = newblocksize;
            }
            unsigned int* fp_v = vendorFPblock + numvendorFPs*this->fpT_stride_ints;
            for (int j = 0; j < this->fpSizeInts; j++) {
                fp_v[j] = char32BitString_to_uint32(fp.substr(32 * j, 32).c_str ());
            }
            numvendorFPs++;
        }
    }
    vendorFile.close();
    unsigned int* counts = (unsigned int*)malloc(numvendorFPs*sizeof(unsigned int));
    memset(counts,0,numvendorFPs*sizeof(unsigned int));
    myScreen_sse4_omp_mag_block(vendorFPblock,numvendorFPs,counts);
    ofstream outfile(this->resultsfile.c_str());
    for (int i = 0; i < numvendorFPs; i++) {
        string id = ids[i];
        outfile << i << " " << id << " " << counts[i] << endl;
    }
    outfile.close();
    free(vendorFPblock);
    free(counts);
}
void NbrCounts::myScreen_hybrid(unsigned int* qfps,const int numqfps,unsigned int* counts){
    unsigned int* magQs = (unsigned int*)malloc(numqfps*4);
    memset(magQs,0,numqfps*4);
    compute_magnitudes(qfps,magQs,numqfps,fpT_stride_ints);
    // This is a dirty hack with hardcoded static load balancing
    // among the GTX 480, GTX 260, and the i7-920
    // approx rates: GTX 480 = 1/16, GTX 260 = 1/70, i7 = 1/80
    // -> approx work fraction = 70% GTX 480; 16% GTX 260, 14% i7
    const int numdbfps = this->numFpsTrue;
    int numqfps_i7 = (numqfps/100)*14;
    numqfps_i7 = ((numqfps_i7 + 63)/64)*64; // Round up to nearest 64
    int numqfp_260 = (numqfps/100)*16;
    int numqfp_480 = numqfps - numqfp_260 - numqfps_i7;
    int numqfp_gpu[2];
    numqfp_gpu[0] = numqfp_480;
    numqfp_gpu[1] = numqfp_260;
    int qfpoffset_i7 = 0;
    int qfpoffset_480 = numqfps_i7;
    int qfpoffset_260 = numqfps_i7 + numqfp_480;
    int qfpoffset_gpu[2] = {qfpoffset_480, qfpoffset_260};
    
    #define OMB_UF 4
    uint64_t* dbfps64 = (uint64_t*)(this->fingerprintsT);
    uint64_t* qfps64 = (uint64_t*)qfps;
    const int fpT_stride_i64 = this->fpT_stride_ints/2;
    unsigned maxfps = (numqfps_i7 > numdbfps) ? numqfps_i7 : numdbfps;
    const int tileSize = 64;
    const int vector_width  = 2;
    const int unroll_factor = OMB_UF;
    assert(numqfps_i7  % tileSize == 0);
    assert(numdbfps % tileSize == 0);
    assert(fpSizeInts % (vector_width*unroll_factor) == 0);
    const int niters = fpSizeInts/(vector_width);
    uint64_t numTiles = ((uint64_t)(maxfps/tileSize))*(maxfps/tileSize);
    unsigned int** allLocalCounts;

    #pragma omp parallel
    {
        unsigned int *d_dfps, *d_qfps, *d_qmags, *d_dmags,*d_counts;
        const int threadIdx = omp_get_thread_num();
        struct gpuData gdata;
        if (threadIdx < 2) {
            gdata.h_dfps   = this->fingerprintsT;
            gdata.h_dmags  = this->magD;
            gdata.h_qfps   = qfps   + qfpoffset_gpu[threadIdx];
            gdata.h_qmags  = magQs  + qfpoffset_gpu[threadIdx];
            gdata.h_counts = counts + qfpoffset_gpu[threadIdx];
            gdata.numdbfps = numdbfps;
            gdata.numqfps  = numqfp_gpu[threadIdx];
            gdata.fpT_stride_ints = this->fpT_stride_ints;
            hybrid_helper_start(gdata,cutoff);
        }
        // Compute with the CPU here
        #pragma omp single
        {
            allLocalCounts = new unsigned int*[omp_get_num_threads()];
        }
        allLocalCounts[threadIdx] = new unsigned int[numqfps_i7]; 
        unsigned int* localCounts = allLocalCounts[threadIdx];
        memset(localCounts,0,numqfps_i7*4);
        #pragma omp for schedule (static,64)
        for (uint64_t tileid = 0; tileid < numTiles; tileid++) {
            // Figure out the coordinates for the tile we're operating on
            uint32_t qtile_idx, dbtile_idx;
            demorton64(&qtile_idx,&dbtile_idx,tileid);
            int qfpBase  =  qtile_idx * tileSize;
            int dbfpBase = dbtile_idx * tileSize;
            if (qfpBase >= numqfps_i7 || dbfpBase >= numdbfps) continue;
            for (int qfpOff = 0; qfpOff < tileSize; qfpOff++) {
                unsigned int qfpIdx = qfpBase + qfpOff;
                unsigned int magQ = magQs[qfpIdx];
                const uint64_t* qfp64 = qfps64 + (qfpIdx)*fpT_stride_i64;
                for (int dbfpOff = 0; dbfpOff < tileSize; dbfpOff++) {
                    unsigned int dbfpIdx = dbfpBase + dbfpOff;
                    const uint64_t* dbfp64 = dbfps64 + (dbfpIdx)*fpT_stride_i64;

                    uint64_t num = 0, num_lo = 0, num_hi = 0, den = 0;
                    uint64_t num_lo2 = 0, num_hi2 = 0;
                    uint64_t q, q2, q3, q4;
                    uint64_t db, db2, db3, db4;
                    uint64_t isct_hi, isct_lo;
                    uint64_t isct_hi2, isct_lo2;
                    for (int iter = 0; iter < niters; iter+=unroll_factor) {
                        q   =  qfp64[iter];
                        db  = dbfp64[iter];
                        q2  =  qfp64[iter+1];
                        db2 = dbfp64[iter+1];
                        q3  =  qfp64[iter+2];
                        db3 = dbfp64[iter+2];
                        q4  =  qfp64[iter+3];
                        db4 = dbfp64[iter+3];
                        isct_lo = q & db;
                        isct_hi = q2 & db2;
                        isct_lo2 = q3 & db3;
                        isct_hi2 = q4 & db4;
                        num_lo += _mm_popcnt_u64(isct_lo);
                        num_hi += _mm_popcnt_u64(isct_hi);
                        num_lo2 += _mm_popcnt_u64(isct_lo2);
                        num_hi2 += _mm_popcnt_u64(isct_hi2);
                    }
                    num = (num_lo + num_hi) + (num_lo2 + num_hi2);
                    den = magQ + magD[dbfpIdx] - num;

                    double tani = (den == 0) ? 0.0 : ((double)num)/den;
                    if (tani >= this->cutoff) localCounts[qfpIdx] += 1;
                }
            }
        }
        #pragma omp single
        {
            // Reduce and free
            for (int t = 0; t < omp_get_num_threads(); t++) {
                unsigned int *lC = allLocalCounts[t];
                for (int i = 0; i < numqfps_i7; i++) counts[i] += lC[i];
                delete[] lC;
            }
            delete[] allLocalCounts;
        }
        // Retrieve GPU results after CPU computation is done
        if (threadIdx < 2) {  // GPUs
            hybrid_helper_finish(gdata);
        }
    }
    free(magQs);
    return;
}
void NbrCounts::run_hybrid(){
    //////////////////// read in vendor fps - these will NOT be end on as in the db fingerprints
    ifstream vendorFile;
    vendorFile.open(this->vendorfile.c_str());
    string line;
    int nbytes_fp_vendor = this->fpSizeInts * sizeof(unsigned int); //32 bit chunks = 4 ints
    string padding (this->fpBitPadLen,'0');
    //const unsigned int* dbfp4 = this->fingerprintsT + (fpIdx+3)*this->fpT_stride_ints;
    unsigned int blocksize = 128;
    unsigned int numvendorFPs = 0;
    unsigned int *vendorFPblock = (unsigned int*)malloc(blocksize*this->fpT_stride_ints*sizeof(unsigned int));
    vector<string> ids;
    for (int i = 0; !vendorFile.eof(); i++) {
        getline(vendorFile, line);
        if (!vendorFile.eof()) {
            stringstream ss(line);
            string id;
            ss >> id;
            ids.push_back(id);
            string fp;
            ss >> fp;
            fp += padding;
            // Check to see if we need to grow the buffer
            if (numvendorFPs == blocksize) {
                unsigned newblocksize = blocksize*2;
                unsigned int* temp = (unsigned int*)malloc(newblocksize*this->fpT_stride_ints*sizeof(unsigned int));
                memcpy(temp,vendorFPblock,blocksize*this->fpT_stride_ints*sizeof(unsigned int));
                free(vendorFPblock);
                vendorFPblock = temp;
                blocksize = newblocksize;
            }
            unsigned int* fp_v = vendorFPblock + numvendorFPs*this->fpT_stride_ints;
            for (int j = 0; j < this->fpSizeInts; j++) {
                fp_v[j] = char32BitString_to_uint32(fp.substr(32 * j, 32).c_str ());
            }
            numvendorFPs++;
        }
    }
    vendorFile.close();
    unsigned int* counts = (unsigned int*)malloc(numvendorFPs*sizeof(unsigned int));
    memset(counts,0,numvendorFPs*sizeof(unsigned int));
    myScreen_hybrid(vendorFPblock,numvendorFPs,counts);
    ofstream outfile(this->resultsfile.c_str());
    for (int i = 0; i < numvendorFPs; i++) {
        string id = ids[i];
        outfile << i << " " << id << " " << counts[i] << endl;
    }
    outfile.close();
    free(vendorFPblock);
    free(counts);
}
